/*
 * $Id: rip2.c,v 1.11 1999/05/10 21:53:38 masaki Exp $
 */

#include <sys/fcntl.h>
#include <mrt.h>
#include <api6.h>
#include <rip.h>


static void
rip2_process_packet_response (gateway_t * gateway, u_char * update, int bytes,
			      int pref)
{
    u_char *cp;
    time_t now;

#define RIP_RTELEN 20

    if ((bytes % RIP_RTELEN) != 0) {
	trace (TR_WARN, RIP->trace, "invalid RTE size %d\n", bytes);
    }

    time (&now);
    cp = update;

    while (cp < update + bytes) {
	int afi, tag, prefixlen, metric;
	struct in_addr addr, mask, nhop;
        prefix_t *prefix;
        dv_attr_t *attr;

	BGP_GET_SHORT (afi, cp);
	BGP_GET_SHORT (tag, cp);
	BGP_GET_ADDR (&addr, cp);
	BGP_GET_ADDR (&mask, cp);
	BGP_GET_ADDR (&nhop, cp);
	BGP_GET_LONG (metric, cp);

	if (afi == 0xffff /* Authentication */ ) {
	    trace (TR_ERROR, RIP->trace,
		   "Authentication not supported (type %d)\n", tag);
	    continue;
	}
	/* IP = 1, IP6 = 2 defined in RFC 1700 p.91 but RIP RFC says 2 for IP */
	if (afi != 2 /* IP AFI */ ) {
	    trace (TR_ERROR, RIP->trace, "unknown afi = %d\n", afi);
	    continue;
	}

	prefixlen = mask2len (&mask, 4);
	prefix = New_Prefix (AF_INET, &addr, prefixlen);

	attr = dv_new_attr (RIP, metric);
	attr->gateway = gateway;
	if (nhop.s_addr == INADDR_ANY) {
	    attr->nexthop = ref_nexthop (gateway);
	}
	else {
	    prefix_t *prefix = New_Prefix (AF_INET, &nhop, 32);
	    attr->nexthop = add_nexthop (prefix, gateway->interface);
	    Deref_Prefix (prefix);
	}
	attr->metric = metric;
	attr->tag = tag;
	attr->utime = now;
	attr->pref = pref;
	dv_update_route (RIP, prefix, attr);
	Deref_Prefix (prefix);
	dv_del_attr (attr);
    }
    if (RIP->changed)
	dv_set_flash_update (RIP);
}


static void
dv_delete_tuple (dv_tuple_t *tuple)
{
    Deref_Prefix (tuple->prefix);
    Delete (tuple);
}


static LINKED_LIST *
rip2_process_packet_request (u_char * update, int bytes)
{
    u_char *cp = update;
    LINKED_LIST *ll_tuple = NULL;

    if ((bytes % RIP_RTELEN) != 0) {
	trace (TR_WARN, RIP->trace, "invalid RTE size %d\n", bytes);
    }

    while (cp < update + bytes) {
	int afi, tag, prefixlen, metric;
	struct in_addr addr, mask, nhop;
        prefix_t *prefix;

	if ((update + bytes) - cp < RIP_RTELEN)
	    break;

	BGP_GET_SHORT (afi, cp);
	BGP_GET_SHORT (tag, cp);
	BGP_GET_ADDR (&addr, cp);
	BGP_GET_ADDR (&mask, cp);
	BGP_GET_ADDR (&nhop, cp);
	BGP_GET_LONG (metric, cp);

	prefixlen = mask2len (&mask, 4);

	/* Address family field in RIP REQUEST must be zero. (RFC1058 3.4.1) 
		"Vladimir V. Ivanov" <vlad@elis.tusur.ru> */
	if (cp == update + bytes && afi == 0 &&
	    metric == RIP_METRIC_INFINITY /* &&
	    prefixlen == 0 && tag == 0 && 
	    addr.s_addr == INADDR_ANY && mask.s_addr == INADDR_ANY &&
	    nhop.s_addr == INADDR_ANY */) {
	    trace (TR_PACKET, RIP->trace,
		   "  whole-table request, responding\n");
	    return (NULL);
	}

	if (afi != 2 /* IP AFI */) {
	    trace (TR_WARN, RIP->trace, "unknown afi = %d\n", afi);
	    continue;
	}

	prefix = New_Prefix (AF_INET, &addr, prefixlen);
	tuple = New (dv_tuple_t);
	tuple->prefix = prefix;
	tuple->attr = NULL;
	tuple->metric = metric;
	if (ll_tuple == NULL)
	    ll_tuple = LL_Create (LL_DestroyFunction, dv_delete_tuple, 0);
    }
    return (ll_tuple);
}


/*
 * start listening for broadcast rip updates
 */
int
rip2_init_listen (prefix_t *prefix)
{
    struct sockaddr_in serv_addr;
    const char zero = 0;
    const int one = 1;
    int sockfd;

    if ((sockfd = open_socket (AF_INET, SOCK_DGRAM, 0)) < 0)
	return (-1);

    reuse_socket (sockfd, 1);
    ip_multicast_loop (sockfd, 0);
    ip_recvdstaddr (sockfd, 1);
    ip_recvttl (sockfd, 1);
    ip_recvinterface (sockfd, 1);

    memset (&serv_addr, 0, sizeof (serv_addr));
    /* check that NOT listening to portmaster! */
    serv_addr.sin_port = htons (RIP->port);
    serv_addr.sin_family = AF_INET;
    if (prefix) {
        serv_addr.sin_addr.s_addr = prefix_tolong (prefix);
    }
    else {
        serv_addr.sin_addr.s_addr = INADDR_ANY;
    }

    if (bind (sockfd,
	      (struct sockaddr *) &serv_addr, sizeof (serv_addr)) < 0) {
	trace (TR_ERROR, RIP->trace, "bind to port %d: %m\n", RIP->port);
	return (-1);
    }
    trace (TR_TRACE, RIP->trace, "listening on port %d at %s\n",
           ntohs (serv_addr.sin_port), (prefix)? prefix_toa (prefix): "*");
    return (sockfd);
}


/*
 * turn on the interface
 */
int
rip2_interface (rip_interface_t *rip_interface, int on)
{
    int ret = 0;
    interface_t *interface = rip_interface->interface;

    if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
	BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
	trace (TR_ERROR, RIP->trace,
               "RIP on interface %s ignored due to no multicast or p2p\n",
                    interface->name);
	return (-1);
    }
    if (BIT_TEST (interface->flags, IFF_MULTICAST))
        ret = join_leave_group (RIP->sockfd, interface, RIP->all_router, on);
    if (ret >= 0 && on)
	rip2_send_request (rip_interface, NULL);
    return (ret);
}


static int
rip2_sendmsgto (rip_interface_t *rip_interface,
		u_char * buffer, int buflen, u_long flag,
		prefix_t *host, int port)
{
    interface_t *interface = rip_interface->interface;
    int sockfd = (rip_interface->sockfd >= 0)?
		    rip_interface->sockfd: RIP->sockfd;

    if (host == NULL) {
	host = RIP->all_routers;
	port = htons (RIP->port);
    }

    return (send_packet (sockfd, buffer, buflen, 0, host, port, interface));
}


/*
 * broadcast request for list of prefixes
 * if NULL, send a request for complete routing table 
 */
void
rip2_send_request (rip_interface_t *rip_interface, LINKED_LIST * ll_prefixes)
{
    u_char buffer[RIP_MAX_PDU], *cp = buffer;

    assert (ll_prefixes == NULL);	/* XXX not yet implemented */

    BGP_PUT_BYTE (RIP_REQUEST, cp);
    BGP_PUT_BYTE (RIP_VERSION, cp);
    BGP_PUT_SHORT (0, cp);	/* unused */
    BGP_PUT_SHORT (0, cp);     /* AF_UNSPEC */
    BGP_PUT_SHORT (0, cp);	/* tag */
    BGP_PUT_NETLONG (0, cp);	/* ip address */
    BGP_PUT_NETLONG (0, cp);	/* subnet mask */
    BGP_PUT_NETLONG (0, cp);	/* next hop */
    BGP_PUT_LONG (RIP_METRIC_INFINITY, cp);

    if (dv_sendmsgto (rip_interface, buffer, cp - buffer, 0, 
			RIP->all_routers, RIP->port, 0) >= 0)
	    trace (TR_TRACE, RIP->trace, "send request on %s\n",
		   rip_interface->interface->name);
}


/*
 * given an interface, broadcast rip routes (according to policy) 
 * we do this every RIP_UPDATE_INTERVAL
 */
rip2_send_update (LINKED_LIST *ll_tuple,
		  interface_t *interface, prefix_t *host, int port)
{
    u_char buffer[RIP_MAX_PDU], *cp = buffer;
    int routes = 0;
    dv_tuple_t *tuple;

    BGP_PUT_BYTE (RIP_RESPONSE, cp);
    BGP_PUT_BYTE (RIP_VERSION, cp);
    BGP_PUT_SHORT (0, cp);	/* unused */

    /* on flashing update, locking should be done outer side
       because of keeping change flags and reset them after sending out */

    LL_Iterate (ll_tuple, tuple) {
        prefix_t *prefix = tuple->prefix;
        dv_attr_t *attr = tuple->attr;
	int metric = tuple->metric;
	struct in_addr addr;

	BGP_PUT_SHORT (2, cp);	/* afi */
	BGP_PUT_SHORT (attr->tag, cp);	/* tag */
	BGP_PUT_ADDR (prefix_tochar (prefix), cp);	/* ip address */
	len2mask (prefix->bitlen, &addr, sizeof addr);
	BGP_PUT_ADDR (&addr, cp);	/* subnet mask */
	BGP_PUT_NETLONG (0, cp);	/* next hop, itself only XXX */
	BGP_PUT_LONG (metric, cp);
	routes++;

	/* see if we have filled up the buffer */
	if (cp - buffer > RIP_MAX_PDU - RIP_RTELEN) {
	    if (rip2_sendmsgto (rip_interface, buffer, cp - buffer, 0,
				host, port) < 0)
		break;

	    if (host) {
		char stmp[INET_ADDRSTRLEN];
		trace (TR_TRACE, RIP->trace, "send %d routes to %s\n",
		       routes, host);
	    }
	    else if (rip_interface)
		trace (TR_TRACE, RIP->trace, "send %d routes on %s\n",
		       routes, rip_interface->interface->name);
	    else
		assert (0);
	    cp = buffer;
	    BGP_PUT_BYTE (RIP_RESPONSE, cp);
	    BGP_PUT_BYTE (RIP_VERSION, cp);
	    BGP_PUT_SHORT (0, cp);	/* unused */
	    routes = 0;
	}

    }

    if (routes > 0) {
	if (rip2_sendmsgto (rip_interface, buffer, cp - buffer, 0,
			    host, port) >= 0) {
	    if (host) {
		trace (TR_TRACE, RIP->trace, "send %d routes to %s\n",
		       routes, host);
	    }
	    else if (rip_interface)
		trace (TR_TRACE, RIP->trace, "send %d routes on %s\n",
		       routes, rip_interface->interface->name);
	    else
		assert (0);
	}
    }
}


/* rip_receive_update
 * read and process and RIP update packet recieved on
 * our interface
 */
void
rip2_receive_update (rip_interface_t *rip_interface)
{
#define RIP_MAX_PDU 512
    u_char buffer[RIP_MAX_PDU], *cp = buffer;
    int n, fromlen, port;
    int command, version, zero;
    struct sockaddr_in from;
    prefix_t *prefix;
    gateway_t *gateway;
    interface_t *interface;
    int sockfd;

    sockfd = (rip_interface && rip_interface->sockfd)? 
		rip_interface->sockfd: RIP->sockfd;
    n = recvmsgfrom (sockfd, buffer, sizeof (buffer), O_NONBLOCK,
                     &prefix, &port, &interface);
    select_enable_fd (sockfd);
    if (n <= 0)
	return;
    assert (prefix);

    if (interface == NULL) {
	trace (TR_WARN, RIP->trace,
	       "discard packet from %a (interface unknown)\n", prefix);
	goto ignore;
    }
    if (rip_interface == NULL) {
        trace (TR_TRACE, RIP->trace,
               "packet from %a at RIP socket %d\n", prefix, sockfd);
        rip_interface = RIP->index2dvi[interface->index];
        assert (rip_interface);
    }
    else if (rip_interface != RIP->index2dvi[interface->index]) {
        trace (TR_ERROR, RIP->trace,
               "confusion: from %a on %s but must be on %s\n", prefix,
               interface->name, rip_interface->interface->name);
        interface = rip_interface->interface;
    }

    if (!BITX_TEST (&RIP->interface_mask, interface->index)) {
        trace (TR_PACKET, RIP->trace,
               "packet from %a on disabled interface %s\n",                     
               prefix, interface->name);         
        goto ignore;
    }

    assert (rip_interface);

    /*
     * get command first for later check
     */
    BGP_GET_BYTE (command, cp);
    BGP_GET_BYTE (version, cp);
    BGP_GET_SHORT (zero, cp);

    if (version != RIP_VERSION) {
	trace (TR_WARN, RIP->trace,
	       "unsupported version %d from %a on %s\n",
	       version, prefix, interface->name);
	goto ignore;
    }

    if (zero) {
	trace (TR_WARN, RIP->trace,
	       "non-zero pad field (value 0x%x) from %a on %s\n",
	       zero, prefix, interface->name);
/* XXX  goto ignore; */
    }

    if (command != RIP_RESPONSE && command != RIP_REQUEST) {
	trace (TR_WARN, RIP->trace,
	       "unsupported command %d from %a on %s, ignore!\n",
	       command, prefix, interface->name);
	goto ignore;
    }

    if (command == RIP_RESPONSE) {

	/* register the gateway */
	gateway = add_gateway (prefix, 0, interface);

	/* don't listen to things broadcast from our own interface */
	if (find_interface_local (prefix)) {
	    trace (TR_PACKET, RIP->trace,
		   "recv ignore own response from %a on %s, ignore!\n",
		   prefix, interface->name);
	    goto ignore;
	}

	/* check if this interface is configured for RIP */
	if (!BITX_TEST (&RIP->interface_mask, interface->index)) {
	    trace (TR_PACKET, RIP->trace, "received response "
		   "from %a on unconfigured interface %s, ignore!\n",
		   gateway->prefix, interface->name);
	    goto ignore;
	}

	if (port != RIP->port) {
	    trace (TR_INFO, RIP->trace,
		   "non-standard source port %d from %a on %s, "
		   "ignored!\n", port,
		   gateway->prefix, interface->name);
	    goto ignore;
	}

	trace (TR_TRACE, RIP->trace,
	       "recv response %d bytes from %a on %s\n",
	       n, gateway->prefix, interface->name);

	if (n - (cp - buffer)) {
	    rip2_process_packet_response (gateway, cp, n - (cp - buffer),
		(rip_interface->default_pref >= 0)? 
			rip_interface->default_pref: RIP->pref);
	}
    }
    else if (command == RIP_REQUEST) {

	/* register the gateway */
	if (port == RIP->port)
	    gateway = add_gateway (prefix, 0, interface);

	/* don't listen to things broadcast from our own interface
	   except for a query */
	if (find_interface_local (prefix) && port == RIP->port) {
	    goto ignore;
	}

	trace (TR_TRACE, RIP->trace,
	       "recv request %d bytes port %d from %a on %s\n",
	       n, port, prefix, interface->name);

	if (n - (cp - buffer) > 0) {
	    /* more than 1 entry */
	    ll_tuple = rip2_process_packet_request (cp, n - (cp - buffer));
	    dv_process_requst (RIP, ll_tuple, rip_interface, prefix, port);
	    trace (TR_PACKET, RIP->trace,
		   "recv request answered to %a port %d on %s\n",
		   prefix, port, interface->name);
	}
	else {
	    trace (TR_WARN, RIP->trace,
		   "recv request no entry from %a on %s, discard it!\n",
		   prefix, interface->name);
	}
    }
  ignore:
    if (prefix)
	Deref_Prefix (prefix);
}
